//decodeimg - console pokewalker image decoder
//(c) 2020 DmitryGR
//GPLv3 license

#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

static const char* colorsFg[] = {"38;5;15", "38;5;253", "38;5;240", "38;5;0"};
static const char* colorsBg[] = {"48;5;15", "48;5;253", "48;5;240", "48;5;0"};



static void escapeSeq(const char *seq, ...)
{
	va_list vl;
	char *str;
	
	va_start(vl, seq);
	putchar(0x1b);
	putchar('[');
	
	while (seq) {
		fputs(seq, stdout);
		seq = va_arg(vl, char*);
	}
	va_end(vl);
}

static void border(void)
{
	putchar('X');
}

static void drawImage(const uint8_t *image, uint32_t w, uint32_t h)
{
	uint32_t r, c;
	
	for (c = 0; c < w + 2; c++)
		border();
	putchar('\n');
	for (r = 0; r < h; r += 2, image += w) {
		border();

		for (c = 0; c < w; c++, image++) {
			
			uint8_t upper = image[0];
			uint8_t lower = r == h - 1 ? 3 : image[w];
			
			upper &= 3;	//2BPP means 2BPP
			lower &= 3;
			
			if (upper == lower) {	//print a space with proper background color
				escapeSeq(colorsBg[upper], "m", NULL);
				putchar(' ');
			}
			else {					//print "upper block" with fore set to uppwer color and back to lower
				escapeSeq(colorsFg[upper], ";", colorsBg[lower], "m", NULL);
				
				//utf8 for upper block
				putchar(0xE2);
				putchar(0x96);
				putchar(0x80);
			}
		}
		escapeSeq("0m", NULL);	//reset
		border();
		putchar('\n');
	}
	for (c = 0; c < w + 2; c++)
		border();
	putchar('\n');
}

static void usage(const char *me)
{
	fprintf(stderr, "USAGE: %s <offset> <num bytes> <img width> < file\n\tvalues can be hex prefixed with '0x' or decimal\n", me);
}

static int32_t getInt(const char *str)
{
	if (!str || !*str)
		return -1;
	
	if (str[0] == '0' && str[1] == 'x')
		return strtoul(str + 2, NULL, 16);
	else
		return strtoul(str, NULL, 10);
}

int main(int argc, char** argv)
{
	int32_t ofst, nBytes, width, i, nWords, height;
	uint16_t *inputData;
	uint8_t *decoded;
	int r, r2, c, c2;
	
	if (argc != 4) {
		usage(argv[0]);
		return 0;
	}
	
	ofst = getInt(argv[1]);
	nBytes = getInt(argv[2]);
	width = getInt(argv[3]);
	
	if (ofst < 0 || nBytes < 0 || width < 0) {
		usage(argv[0]);
		return 0;
	}
	
	if ((nBytes % 2) || ((nBytes / 2)% width)) {
		fprintf(stderr, "provided width %u does not mesh with provided byte length %u\n", width, nBytes);
		usage(argv[0]);
		return 0;
	}
	
	//skip start
	while (ofst-- && (c = getchar()) != EOF);
	if (c == EOF) {
		fprintf(stderr, "input file too short for the given offset\n");
		usage(argv[0]);
		return 0;
	}
	
	//alloc space for and read the data
	nWords = nBytes / 2;
	inputData = malloc(nBytes * sizeof(uint16_t));
	for (i = 0; i < nWords; i++) {
	
		if ((c = getchar()) == EOF || (c2 = getchar()) == EOF) {
			fprintf(stderr, "input file too short for the given length\n");
			usage(argv[0]);
			return 0;
		}
		inputData[i] = (c << 8) + c2;
	}
	
	//alloc space for the image
	height = nWords * 8 / width;
	decoded = malloc(width * height * sizeof(uint8_t));
	
	//decode
	for (r = 0; r < height; r += 8) {
		for (c = 0; c < width; c++, inputData++) {
			for (r2 = 0; r2 < 8; r2++) {
				switch ((*inputData >> r2) & 0x0101) {
					case 0x000:	//white
						decoded[(r + r2) * width + c] = 0b00;
						break;
					case 0x001:	//dark grey
						decoded[(r + r2) * width + c] = 0b01;
						break;
					case 0x100:	//light grey
						decoded[(r + r2) * width + c] = 0b10;
						break;
					case 0x101:	//black
						decoded[(r + r2) * width + c] = 0b11;
						break;
				}
			}
		}
	}
	
	fprintf(stderr, "image %u x %u\n", width, height);
	drawImage(decoded, width, height);
	
	return 0;
}
